"""
Implements single text negotiation mechanisms.
"""
from __future__ import annotations

import math
import random
import time
from copy import deepcopy

from attr import define

from .mechanisms import Mechanism, MechanismRoundResult, MechanismState
from .outcomes import Outcome

__all__ = ["VetoSTMechanism", "HillClimbingSTMechanism"]


@define
class STState(MechanismState):
    """Defines extra values to keep in the mechanism state. This is accessible to all negotiators"""

    current_offer: Outcome | None = None
    new_offer: Outcome | None = None


class VetoSTMechanism(Mechanism):
    """Base class for all single text mechanisms

    Args:
        *args: positional arguments to be passed to the base Mechanism
        **kwargs: keyword arguments to be passed to the base Mechanism
        initial_outcome: initial outcome. If None, it will be selected by `next_outcome` which by default will choose it
                         randomly.
        initial_responses: Initial set of responses.

    Remarks:

        - initial_responses is only of value when the number of negotiators that will join the negotiation is less then
          or equal to its length. By default it is not used for anything. Nevertheless, it is here because
          `next_outcome` may decide to use it with the `initial_outcome`


    """

    def __init__(
        self,
        *args,
        epsilon: float = 1e-6,
        initial_outcome=None,
        initial_responses: tuple[bool] = tuple(),
        **kwargs,
    ):
        kwargs["state_factory"] = STState
        super().__init__(*args, **kwargs)

        self.add_requirements(
            {"compare-binary": True}
        )  # assert that all agents must have compare-binary capability
        self._current_state.current_offer = initial_outcome
        """The current offer"""
        self.initial_outcome = deepcopy(initial_outcome)
        """The initial offer"""
        self.last_responses = list(initial_responses)
        """The responses of all negotiators for the last offer"""
        self.initial_responses = deepcopy(self.last_responses)
        """The initial set of responses. See the remarks of this class to understand its role."""
        self.epsilon = epsilon
        self._current_state.new_offer = initial_outcome
        """The new offer generated in this step"""

    def next_outcome(self, outcome: Outcome | None) -> Outcome | None:
        """Generate the next outcome given some outcome.

        Args:
             outcome: The current outcome

        Returns:
            a new outcome or None to end the mechanism run

        """
        return self.random_outcomes(1)[0]

    def round(self) -> MechanismRoundResult:
        """Single round of the protocol"""
        new_offer = self.next_outcome(self._current_state.current_offer)
        responses = []

        for neg in self.negotiators:
            strt = time.perf_counter()
            responses.append(
                neg.is_better(new_offer, self._current_state.current_offer) is not False
            )
            if time.perf_counter() - strt > self.nmi.step_time_limit:
                return MechanismRoundResult(broken=False, timedout=True, agreement=None)

        self.last_responses = responses
        self._current_state.new_offer = new_offer
        if all(responses):
            self._current_state.current_offer = new_offer

        return MechanismRoundResult(broken=False, timedout=False, agreement=None)

    def on_negotiation_end(self) -> None:
        """Used to pass the final offer for agreement between all negotiators"""
        if self._current_state.current_offer is not None and all(
            neg.is_acceptable_as_agreement(self._current_state.current_offer)
            for neg in self.negotiators
        ):
            self._current_state.agreement = self._current_state.current_offer

        super().on_negotiation_end()

    def plot(
        self,
        visible_negotiators: tuple[int, int] | tuple[str, str] = (0, 1),
        show_all_offers=False,
        **kwargs,
    ):
        import matplotlib.gridspec as gridspec
        import matplotlib.pyplot as plt
        import pandas as pd

        if len(self.negotiators) < 2:
            print("Cannot visualize negotiations with more less than 2 negotiators")
            return
        if len(visible_negotiators) > 2:
            print("Cannot visualize more than 2 agents")
            return
        if isinstance(visible_negotiators[0], str):
            tmp = []
            for _ in visible_negotiators:
                for n in self.negotiators:
                    if n.id == _:
                        tmp.append(n)
        else:
            visible_negotiators = [
                self.negotiators[visible_negotiators[0]],
                self.negotiators[visible_negotiators[1]],
            ]
        # indx = dict(zip([_.id for _ in self.negotiators], range(len(self.negotiators))))
        history = []
        for state in self.history:
            offer = state.new_offer if show_all_offers else state.current_offer
            history.append(
                {
                    "current_offer": offer,
                    "relative_time": state.relative_time,
                    "step": state.step,
                    "u0": visible_negotiators[0].ufun(offer),
                    "u1": visible_negotiators[1].ufun(offer),
                }
            )
        history = pd.DataFrame(data=history)
        has_history = len(history) > 0
        has_front = 1
        # n_negotiators = len(self.negotiators)
        n_agents = len(visible_negotiators)
        ufuns = self._get_preferencess()
        outcomes = self.outcomes
        utils = [tuple(f(o) for f in ufuns) for o in outcomes]
        agent_names = [a.name for a in visible_negotiators]
        frontier, frontier_outcome = self.pareto_frontier(sort_by_welfare=True)
        frontier_indices = [
            i
            for i, _ in enumerate(frontier)
            if _[0] is not None
            and _[0] > float("-inf")
            and _[1] is not None
            and _[1] > float("-inf")
        ]
        frontier = [frontier[i] for i in frontier_indices]
        frontier_outcome = [frontier_outcome[i] for i in frontier_indices]
        # frontier_outcome_indices = [outcomes.index(_) for _ in frontier_outcome]

        fig_util = plt.figure()
        gs_util = gridspec.GridSpec(n_agents, has_front + 1)
        axs_util = []

        for a in range(n_agents):
            if a == 0:
                axs_util.append(fig_util.add_subplot(gs_util[a, has_front]))
            else:
                axs_util.append(
                    fig_util.add_subplot(gs_util[a, has_front], sharex=axs_util[0])
                )
            axs_util[-1].set_ylabel(agent_names[a])
        for a, au in enumerate(axs_util):
            if au is None:
                break
            if has_history:
                h = history.loc[:, ["step", "current_offer", "u0", "u1"]]
                h["utility"] = h[f"u{a}"]
                au.plot(h.step, h.utility)
                au.set_ylim(0.0, 1.0)

        if has_front:
            axu = fig_util.add_subplot(gs_util[:, 0])
            axu.scatter(
                [_[0] for _ in utils],
                [_[1] for _ in utils],
                label="outcomes",
                color="gray",
                marker="s",
                s=20,
            )
            f1, f2 = [_[0] for _ in frontier], [_[1] for _ in frontier]
            axu.scatter(f1, f2, label="frontier", color="red", marker="x")
            # axu.legend()
            axu.set_xlabel(agent_names[0] + " utility")
            axu.set_ylabel(agent_names[1] + " utility")
            if self.agreement is not None:
                pareto_distance = 1e9
                cu = (ufuns[0](self.agreement), ufuns[1](self.agreement))
                for pu in frontier:
                    dist = math.sqrt((pu[0] - cu[0]) ** 2 + (pu[1] - cu[1]) ** 2)
                    if dist < pareto_distance:
                        pareto_distance = dist
                axu.text(
                    0.05,
                    0.05,
                    f"Pareto-distance={pareto_distance:5.2}",
                    verticalalignment="top",
                    transform=axu.transAxes,
                )

            if has_history:
                h = history.loc[:, ["step", "current_offer", "u0", "u1"]]
                axu.scatter(h.u0, h.u1, color="green", label=f"Mediator's Offer")
                axu.scatter(
                    [frontier[0][0]],
                    [frontier[0][1]],
                    color="blue",
                    label=f"Max Welfare",
                )
                # axu.annotate(
                #     "Max. Welfare",
                #     xy=frontier[0],  # theta, radius
                #     xytext=(
                #         frontier[0][0] + 0.1,
                #         frontier[0][1] + 0.02,
                #     ),  # fraction, fraction
                #     arrowprops=dict(facecolor="black", shrink=0.05),
                #     horizontalalignment="left",
                #     verticalalignment="bottom",
                # )
            if self.state.agreement is not None:
                axu.scatter(
                    [ufuns[0](self.state.agreement)],
                    [ufuns[1](self.state.agreement)],
                    color="black",
                    marker="*",
                    s=120,
                    label="Agreement",
                )

        fig_util.show()

    @property
    def current_offer(self):
        return self.state.current_offer


class HillClimbingSTMechanism(VetoSTMechanism):
    """A single text mechanism that use hill climbing

    Args:
        *args: positional arguments to be passed to the base Mechanism
        **kwargs: keyword arguments to be passed to the base Mechanism
    """

    def neighbors(self, outcome: Outcome) -> list[Outcome]:
        """Returns all neighbors

        Neighbor is an outcome that differs any one of the issues from the original outcome.
        """

        neighbors = []
        for i, issue in enumerate(self.issues):
            values = []
            if isinstance(issue.values, list):
                values = issue.values
            if isinstance(issue.values, int):
                values = [
                    max(0, outcome[i] - 1),
                    min(outcome[i] + 1, issue.values),
                ]
            if isinstance(issue.values, tuple):
                delta = random.random() * (issue.values[0] - issue.values[0])
                values.append(max(issue.values[0], outcome[i] - delta))
                values.append(min(outcome[i] + delta, issue.values[0]))

            for value in values:
                neighbor = list(deepcopy(outcome))
                if neighbor[i] == value:
                    continue
                neighbor[i] = value
                neighbors.append(tuple(neighbor))

        return neighbors

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for issue in self.issues:
            if issue.is_discrete() is False:
                raise ValueError("This mechanism assume discrete issues")

        if self.initial_outcome is None:
            self.initial_outcome = self.random_outcomes(1)[0]

        self._current_state.current_offer = self.initial_outcome
        self.possible_offers = self.neighbors(self._current_state.current_offer)

    def next_outcome(self, outcome: Outcome) -> Outcome | None:
        """Generate the next outcome given some outcome.

        Args:
             outcome: The current outcome

        Returns:
            a new outcome or None to end the mechanism run

        """

        if len(self.possible_offers) == 0:
            return None
        return self.possible_offers.pop(
            random.randint(0, len(self.possible_offers)) - 1
        )

    def round(self) -> MechanismRoundResult:
        """Single round of the protocol"""

        new_offer = self.next_outcome(self._current_state.current_offer)
        if new_offer is None:
            return MechanismRoundResult(
                broken=False,
                timedout=False,
                agreement=self._current_state.current_offer,
            )

        responses = []
        for neg in self.negotiators:
            strt = time.perf_counter()
            responses.append(
                neg.is_better(new_offer, self._current_state.current_offer) is not False
            )
            if time.perf_counter() - strt > self.nmi.step_time_limit:
                return MechanismRoundResult(broken=False, timedout=True, agreement=None)

        self.last_responses = responses

        if all(responses):
            self._current_state.current_offer = new_offer
            self.possible_offers = self.neighbors(self._current_state.current_offer)

        return MechanismRoundResult(broken=False, timedout=False, agreement=None)

    @property
    def current_offer(self):
        return self.state.current_offer



from __future__ import annotations

import os

import pkg_resources
import pytest

from negmas import AspirationNegotiator, load_genius_domain_from_folder


@pytest.fixture
def scenarios_folder():
    return pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/scenarios"
    )


def test_importing_file_without_exceptions(scenarios_folder):
    folder_name = scenarios_folder + "/other/S-1NIKFRT-1"
    load_genius_domain_from_folder(folder_name, n_discretization=10)


def test_simple_run_with_aspiration_agents():
    file_name = pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/Laptop"
    )
    assert os.path.exists(file_name)
    domain = load_genius_domain_from_folder(file_name).to_single_issue()
    mechanism = domain.make_session(AspirationNegotiator, n_steps=100, time_limit=30)
    assert mechanism is not None
    mechanism.run()



from __future__ import annotations

import socket
import sys

__all__ = [
    "DEFAULT_JAVA_PORT",
    "DEFAULT_PYTHON_PORT",
    "DEFAULT_GENIUS_NEGOTIATOR_TIMEOUT",
    "ANY_JAVA_PORT",
    "RANDOM_JAVA_PORT",
    "get_free_tcp_port",
]
DEFAULT_JAVA_PORT = 25337
DEFAULT_PYTHON_PORT = 25338
RANDOM_JAVA_PORT = 0
ANY_JAVA_PORT = -1
DEFAULT_GENIUS_NEGOTIATOR_TIMEOUT = sys.maxsize


def get_free_tcp_port():
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(("", 0))
    addr, port = tcp.getsockname()
    tcp.close()
    return port



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from negmas.common import PreferencesChange
from negmas.helpers import snake_case
from negmas.helpers.types import get_full_type_name
from negmas.outcomes import Outcome
from negmas.outcomes.common import check_one_at_most, os_or_none
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, deserialize, serialize
from negmas.types import NamedObject

__all__ = ["Preferences"]

if TYPE_CHECKING:
    from negmas import Rational
    from negmas.outcomes.base_issue import Issue
    from negmas.outcomes.common import Outcome
    from negmas.outcomes.protocols import OutcomeSpace


class Preferences(NamedObject, ABC):
    """
    Base class for all preferences.

    Args:
        outcome_space: The outcome-space over which the preferences are defined
    """

    outcome_space: OutcomeSpace | None
    reserved_outcome: Outcome
    owner: Rational | None = None

    def __init__(
        self,
        *args,
        outcome_space: OutcomeSpace | None = None,
        issues: tuple[Issue] | None = None,
        outcomes: tuple[Outcome] | int | None = None,
        reserved_outcome: Outcome | None = None,
        **kwargs,
    ) -> None:
        self.owner = None
        check_one_at_most(outcome_space, issues, outcomes)
        super().__init__(*args, **kwargs)
        self.outcome_space = os_or_none(outcome_space, issues, outcomes)
        self.reserved_outcome = reserved_outcome  # type: ignore
        self._changes: list[PreferencesChange] = []

    @abstractmethod
    def is_volatile(self):
        """
        Does the utiltiy of an outcome depend on factors outside the negotiation?


        Remarks:
            - A volatile preferences is one that can change even for the same mechanism state due to outside influence
        """

    @abstractmethod
    def is_not_worse(self, first: Outcome | None, second: Outcome | None) -> bool:
        """
        Is `first` at least as good as `second`
        """
        ...

    @abstractmethod
    def is_session_dependent(self):
        """
        Does the utiltiy of an outcome depend on the `NegotiatorMechanismInterface`?
        """

    @abstractmethod
    def is_state_dependent(self):
        """
        Does the utiltiy of an outcome depend on the negotiation state?
        """

    def to_dict(self) -> dict[str, Any]:
        d = {PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self))}
        return dict(
            **d,
            outcome_space=serialize(self.outcome_space),
            reserved_outcome=self.reserved_outcome,
            name=self.name,
            id=self.id,
        )

    @classmethod
    def from_dict(cls, d):
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        d["outcome_space"] = deserialize(d.get("outcome_space", None))
        return cls(**d)

    def is_stationary(self) -> bool:
        """Are the preferences stationary (i.e. repeated calls return the same value for any preferences comparion or evaluaton method)?"""
        return (
            not self.is_state_dependent()
            and not self.is_volatile()
            and not self.is_session_dependent()
        )

    def changes(self) -> list[PreferencesChange]:
        if self.is_stationary():
            return []
        return [PreferencesChange()]

    @property
    def type(self) -> str:
        """Returns the utility_function type.

        Each class inheriting from this ``UtilityFunction`` class will have its own type. The default type is the empty
        string.

        Examples:
            >>> from negmas.preferences import *
            >>> from negmas.outcomes import make_issue
            >>> print(LinearAdditiveUtilityFunction((lambda x:x, lambda x:x), issues=[make_issue((0, 1), (0, 1))]).type)
            linear_additive
            >>> print(MappingUtilityFunction([lambda x: x], issues=[make_issue((0.0, 1.0))]).type)
            mapping

        Returns:
            str: utility_function type
        """
        return snake_case(
            self.__class__.__name__.replace("Function", "").replace("Utility", "")
        )

    @property
    def base_type(self) -> str:
        """Returns the utility_function base type ignoring discounting and similar wrappings."""
        from .discounted import DiscountedUtilityFunction

        u = self
        while isinstance(u, DiscountedUtilityFunction):
            u = u.ufun
        return self.type

    def is_better(self, first: Outcome | None, second: Outcome | None) -> bool:
        """
        Compares two offers using the `ufun` returning whether the first is strictly better than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared

        Remarks:

            - Should raise `ValueError` if the comparison cannot be done

        """
        return self.is_not_worse(first, second) and not self.is_not_worse(second, first)

    def is_equivalent(self, first: Outcome | None, second: Outcome | None) -> bool:
        """
        Compares two offers using the `ufun` returning whether the first is strictly equivelent than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared

        Remarks:

            - Should raise `ValueError` if the comparison cannot be done

        """
        return self.is_not_worse(first, second) and self.is_not_worse(second, first)

    def is_not_better(self, first: Outcome, second: Outcome | None) -> bool:
        """
        Compares two offers using the `ufun` returning whether the first is worse or equivalent than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared

        Remarks:

            - Should raise `ValueError` if the comparison cannot be done

        """
        return self.is_not_worse(second, first)

    def is_worse(self, first: Outcome | None, second: Outcome | None) -> bool:
        """
        Compares two offers using the `ufun` returning whether the first is strictly worse than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared

        Remarks:

            - Should raise `ValueError` if the comparison cannot be done

        """
        return not self.is_not_worse(first, second)



from __future__ import annotations

"""Defines basic config for NEGMAS"""
import json
import pathlib

__all__ = [
    "NEGMAS_CONFIG",
    "CONFIG_KEY_JNEGMAS_JAR",
    "CONFIG_KEY_GENIUS_BRIDGE_JAR",
]

NEGMAS_DEFAULT_PATH = "~/negmas/config.json"
"""Default path for NegMAS configurations"""

CONFIG_KEY_JNEGMAS_JAR = "jnegmas_jar"
"""Key name for the JNegMAS jar in `NEGMAS_CONFIG`"""

CONFIG_KEY_GENIUS_BRIDGE_JAR = "genius_bridge_jar"
"""Key name for the Genius bridge jar in `NEGMAS_CONFIG`"""

NEGMAS_CONFIG = {
    CONFIG_KEY_JNEGMAS_JAR: f"{NEGMAS_DEFAULT_PATH}/{CONFIG_KEY_JNEGMAS_JAR}",
    CONFIG_KEY_GENIUS_BRIDGE_JAR: f"{NEGMAS_DEFAULT_PATH}/{CONFIG_KEY_GENIUS_BRIDGE_JAR}",
}
# loading config file if any
__conf_path = pathlib.Path(NEGMAS_DEFAULT_PATH).expanduser().absolute()

if __conf_path.exists():
    with open(__conf_path) as f:
        NEGMAS_CONFIG = json.load(f)



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, Iterable

from negmas.common import (
    MechanismState,
    NegotiatorMechanismInterface,
    PreferencesChange,
)

from .components.component import Component
from .negotiator import Negotiator

if TYPE_CHECKING:
    from negmas.preferences import BaseUtilityFunction, Preferences

__all__ = [
    "ModularNegotiator",
]


class ModularNegotiator(Negotiator):
    """
    A generic modular negotiator that can combine multiple negotiation `Component` s.


    This class simply holds a list of components and call them on every event.
    """

    def insert_component(
        self,
        component: Component,
        name: str | None = None,
        index: int = -1,
        override: bool = False,
    ) -> None:
        """
        Adds a component at the given index. If a negative number is given, appends at the end
        """
        if (
            component.negotiator is not None
            and id(component.negotiator) != id(self)
            and not override
        ):
            raise ValueError(
                f"Component {component} already has as parent {component.negotiator.id} that is not the same as this negotiator {self.id}. Cannot add it. To override pass `overrride=True`"
            )
        component.set_negotiator(self)
        if index < 0:
            index = len(self._components)
        self._components.insert(index, component)
        if not name:
            name = str(index)
        self.__component_map[name] = len(self._components)

    def __init__(
        self,
        *args,
        components: Iterable[Component],
        component_names: Iterable[str] | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._components: list[Component] = []
        self.__component_map: dict[str, int] = dict()
        for c, n in zip(
            components, component_names if component_names else itertools.repeat(None)
        ):
            self.insert_component(c, name=n)

    @property
    def components(self) -> tuple[Component, ...]:
        return tuple(self._components)

    def remove_component_at(self, index: int) -> None:
        """
        Removes the component at the givne index.
        """
        self._components = self._components[:index] + self._components[index + 1 :]

    def remove_component(self, name: str) -> None:
        """
        Removes the component with the given name
        """
        self.remove_component_at(self.__component_map[name])

    def on_preferences_changed(self, changes: list[PreferencesChange]):
        """
        Called to inform the component that the ufun has changed and the kinds of change that happened.
        """
        for c in self._components:
            c.on_preferences_changed(changes)

    def join(
        self,
        nmi: NegotiatorMechanismInterface,
        state: MechanismState,
        *,
        preferences: Preferences | None = None,
        ufun: BaseUtilityFunction | None = None,
        role: str = "negotiator",
    ) -> bool:
        if not all(
            _.can_join(nmi, state, preferences=preferences, ufun=ufun, role=role)
            for _ in self._components
        ):
            return False
        joined = super().join(nmi, state, preferences=preferences, ufun=ufun, role=role)
        if not joined:
            return False
        for c in self._components:
            c.after_join(nmi)
        return joined

    def on_negotiation_start(self, state: MechanismState) -> None:
        """
        A call back called at each negotiation start
        """
        for c in self._components:
            c.on_negotiation_start(state)

    def on_round_start(self, state: MechanismState) -> None:
        """
        A call back called at each negotiation round start
        """
        for c in self._components:
            c.on_round_start(state)

    def on_round_end(self, state: MechanismState) -> None:
        """
        A call back called at each negotiation round end
        """
        for c in self._components:
            c.on_round_end(state)

    def on_leave(self, state: MechanismState) -> None:
        """
        A call back called after leaving a negotiation.
        """
        for c in self._components:
            c.on_leave(state)

    def on_negotiation_end(self, state: MechanismState) -> None:
        """
        A call back called at each negotiation end
        """
        for c in self._components:
            c.on_negotiation_end(state)

    def on_mechanism_error(self, state: MechanismState) -> None:
        """
        A call back called whenever an error happens in the mechanism. The error and its explanation are accessible in
        `state`
        """
        for c in self._components:
            c.on_mechanism_error(state)



from __future__ import annotations

import math
import pathlib
import uuid
from typing import TYPE_CHECKING, Callable

from negmas.helpers.misc import make_callable
from negmas.outcomes.base_issue import Issue
from negmas.outcomes.common import Outcome, os_or_none
from negmas.outcomes.outcome_space import make_os
from negmas.outcomes.protocols import OutcomeSpace
from negmas.preferences.crisp_ufun import UtilityFunction
from negmas.preferences.ops import nash_point, pareto_frontier
from negmas.sao.mechanism import TraceElement

if TYPE_CHECKING:
    from matplotlib.axes import Axes

__all__ = ["plot_offer_utilities", "plot_mechanism_run", "plot_2dutils"]

ALL_MARKERS = ["s", "o", "v", "^", "<", ">", "p", "P", "h", "H", "1", "2", "3", "4"]
PROPOSALS_ALPHA = 0.9
AGREEMENT_ALPHA = 0.9
NASH_ALPHA = 0.6
RESERVED_ALPHA = 0.08
WELFARE_ALPHA = 0.6
AGREEMENT_SCALE = 10
NASH_SCALE = 4
WELFARE_SCALE = 2
OUTCOMES_SCALE = 0.5
PARETO_SCALE = 1.5


def get_cmap(n, name="jet"):
    """Returns a function that maps each index in 0, 1, ..., n-1 to a distinct
    RGB color; the keyword argument name must be a standard mpl colormap name."""
    import matplotlib.pyplot as plt

    return plt.cm.get_cmap(name, n)


def make_colors_and_markers(colors, markers, n: int, colormap="jet"):
    if not colors:
        cmap = get_cmap(n, colormap)
        colors = [cmap(i) for i in range(n)]
    if not markers:
        markers = [ALL_MARKERS[i % len(ALL_MARKERS)] for i in range(n)]
    return colors, markers


def plot_offer_utilities(
    trace: list[TraceElement],
    negotiator: str,
    plotting_ufuns: list[UtilityFunction],
    plotting_negotiators: list[str],
    ignore_none_offers: bool = True,
    name_map: dict[str, str] | Callable[[str], str] | None = None,
    colors: list | None = None,
    markers: list | None = None,
    colormap: str = "jet",
    ax: Axes | None = None,  # type: ignore
    sharey=False,
    xdim: str = "relative_time",
    ylimits: tuple[float, float] | None = None,
    show_legend=True,
    show_x_label=True,
    ignore_markers_limit=50,
    show_reserved=True,
):
    import matplotlib.pyplot as plt

    map_ = make_callable(name_map)
    if ax is None:
        _, ax = plt.subplots()
    ax: Axes
    one_y = True
    axes = [ax] * len(plotting_negotiators)
    if not sharey and len(plotting_negotiators) == 2:
        axes = [ax, ax.twinx()]
        one_y = False

    colors, markers = make_colors_and_markers(
        colors, markers, len(plotting_negotiators), colormap
    )

    if xdim.startswith("step") or xdim.startswith("round"):
        trace_info = [(_.offer, _.step) for _ in trace if _.negotiator == negotiator]
    elif xdim.startswith("time") or xdim.startswith("real"):
        trace_info = [(_.offer, _.time) for _ in trace if _.negotiator == negotiator]
    else:
        trace_info = [
            (_.offer, _.relative_time) for _ in trace if _.negotiator == negotiator
        ]
    x = [_[-1] for _ in trace_info]
    for i, (u, neg, a) in enumerate(zip(plotting_ufuns, plotting_negotiators, axes)):
        name = map_(neg)
        y = [u(_[0]) for _ in trace_info]
        reserved = None
        if show_reserved:
            r = u.reserved_value
            if r is not None and math.isfinite(r):
                reserved = [r] * len(y)
        if not ignore_none_offers:
            xx = x
        else:
            good_indices = [i for i, _ in enumerate(y) if _ is not None]
            xx = [x[_] for _ in good_indices]
            y = [y[_] for _ in good_indices]
        a.plot(
            xx,
            y,
            label=f"{name} ({i})",
            color=colors[i % len(colors)],
            linestyle="solid" if neg == negotiator else "dotted",
            linewidth=2 if neg == negotiator else 1,
            marker=markers[i % len(markers)]
            if len(trace_info) < ignore_markers_limit
            else None,
        )
        if reserved:
            a.plot(
                xx,
                reserved,
                # label=f"{name} ({i})",
                color=colors[i % len(colors)],
                linestyle="solid" if neg == negotiator else "dotted",
                linewidth=0.5,
                marker=None,
            )
        if ylimits is not None:
            a.set_ylim(ylimits)
        a.set_ylabel(f"{name} ({i}) utility" if not one_y else "utility")
        if show_legend and len(plotting_negotiators) == 2:
            a.legend(
                loc=f"upper {'left' if not i else 'right'}", bbox_to_anchor=(i, 1.2)
            )
    if show_legend and len(plotting_negotiators) != 2:
        ax.legend(
            bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
            loc="lower left",
            ncol=2,
            mode="expand",
            borderaxespad=0.0,
        )
    if show_x_label:
        ax.set_xlabel(xdim)


def plot_2dutils(
    trace: list[TraceElement],
    plotting_ufuns: list[UtilityFunction],
    plotting_negotiators: list[str],
    offering_negotiators: list[str] | None = None,
    agreement: Outcome | None = None,
    outcome_space: OutcomeSpace | None = None,
    issues: list[Issue] | None = None,
    outcomes: list[Outcome] | None = None,
    with_lines: bool = True,
    show_annotations: bool = True,
    show_agreement: bool = False,
    show_pareto_distance: bool = True,
    show_nash_distance: bool = True,
    show_reserved: bool = True,
    end_reason: str | None = None,
    last_negotiator: str | None = None,
    name_map: dict[str, str] | Callable[[str], str] | None = None,
    colors: list | None = None,
    markers: list[str] | None = None,
    colormap: str = "jet",
    ax: Axes | None = None,  # type: ignore
):
    import matplotlib.patches as mpatches
    import matplotlib.pyplot as plt

    if ax is None:
        _, ax = plt.subplots()
    ax: Axes
    map_ = make_callable(name_map)
    if not outcomes:
        outcome_space = os_or_none(outcome_space, issues, outcomes)
        if outcome_space:
            outcomes = list(outcome_space.enumerate_or_sample(10, 1000))
    if not outcomes:
        outcomes = list({_.offer for _ in trace})
    if not outcome_space:
        outcome_space = make_os(issues=issues, outcomes=outcomes)
    if not offering_negotiators:
        offering_negotiators = list({_.negotiator for _ in trace})

    utils = [tuple(f(o) for f in plotting_ufuns) for o in outcomes]
    max(_[0] for _ in utils) - min(_[0] for _ in utils)
    yrange = max(_[1] for _ in utils) - min(_[1] for _ in utils)
    frontier, frontier_outcome = pareto_frontier(
        ufuns=plotting_ufuns,
        issues=issues,
        outcomes=outcomes if not issues else None,
        sort_by_welfare=True,
    )
    nash, _ = nash_point(plotting_ufuns, frontier, outcome_space=outcome_space)
    if not nash:
        show_nash_distance = False
    frontier_indices = [
        i
        for i, _ in enumerate(frontier)
        if _[0] is not None
        and _[0] > float("-inf")
        and _[1] is not None
        and _[1] > float("-inf")
    ]
    frontier = [frontier[i] for i in frontier_indices]
    frontier_outcome = [frontier_outcome[i] for i in frontier_indices]

    colors, markers = make_colors_and_markers(
        colors, markers, len(offering_negotiators), colormap
    )

    agreement_utility = tuple(u(agreement) for u in plotting_ufuns)
    unknown_agreement_utility = None in agreement_utility
    if unknown_agreement_utility:
        show_pareto_distance = show_nash_distance = False
    default_marker_size = plt.rcParams.get("lines.markersize", 20) ** 2
    ax.scatter(
        [_[0] for _ in utils],
        [_[1] for _ in utils],
        color="gray",
        marker=".",
        s=int(default_marker_size * OUTCOMES_SCALE),
    )
    agent_names = [map_(_) for _ in plotting_negotiators]
    pareto_distance = float("inf")
    nash_distance = float("inf")
    f1, f2 = [_[0] for _ in frontier], [_[1] for _ in frontier]
    ax.scatter(
        f1, f2, color="red", marker="x", s=int(default_marker_size * PARETO_SCALE)
    )
    ax.set_xlabel(agent_names[0] + f"(0) utility")
    ax.set_ylabel(agent_names[1] + f"(1) utility")
    cu = agreement_utility
    if not unknown_agreement_utility:
        if nash:
            nash_distance = math.sqrt((nash[0] - cu[0]) ** 2 + (nash[1] - cu[1]) ** 2)
        for pu in frontier:
            dist = math.sqrt((pu[0] - cu[0]) ** 2 + (pu[1] - cu[1]) ** 2)
            if dist < pareto_distance:
                pareto_distance = dist
    txt = ""
    if show_agreement:
        txt += f"Agreement:{agreement}\n"
    if show_pareto_distance and agreement is not None:
        txt += f"Pareto-distance={pareto_distance:5.2}\n"
    if show_nash_distance and agreement is not None:
        txt += f"Nash-distance={nash_distance:5.2}\n"
    if end_reason:
        txt += f"{end_reason}\n"
    if last_negotiator:
        txt += f"Last: {last_negotiator}\n"

    ax.text(
        0.05,
        0.05,
        txt,
        verticalalignment="bottom",
        transform=ax.transAxes,
    )

    if show_reserved:
        ranges = [
            plotting_ufuns[_].minmax(outcome_space=outcome_space)
            for _ in range(len(plotting_ufuns))
        ]
        for i, (mn, mx) in enumerate(ranges):
            if any(_ is None or not math.isfinite(_) for _ in (mn, mx)):
                x = []
                for a, neg in enumerate(offering_negotiators):
                    negtrace = [_ for _ in trace if _.negotiator == neg]
                    x += [plotting_ufuns[i](_.offer) for _ in negtrace]
                ranges[i] = (min(x), max(x))
        for i, (mn, mx) in enumerate(ranges):
            r = plotting_ufuns[i].reserved_value
            if r is None or not math.isfinite(r):
                r = mn
            if i == 0:
                pt = (r, ranges[1 - i][0])
                width = mx - r
                height = ranges[1 - i][1] - ranges[1 - i][0]
            else:
                pt = (ranges[1 - i][0], r)
                height = mx - r
                width = ranges[1 - i][1] - ranges[1 - i][0]

            ax.add_patch(
                mpatches.Rectangle(
                    pt,
                    width=width,
                    height=height,
                    fill=True,
                    color=colors[i % len(colors)],
                    alpha=RESERVED_ALPHA,
                    linewidth=0,
                )
            )
    for a, neg in enumerate(offering_negotiators):
        negtrace = [_ for _ in trace if _.negotiator == neg]
        x = [plotting_ufuns[0](_.offer) for _ in negtrace]
        y = [plotting_ufuns[1](_.offer) for _ in negtrace]
        (ax.scatter if not with_lines else ax.plot)(
            x,
            y,
            color=colors[a % len(colors)],
            alpha=PROPOSALS_ALPHA,
            label=f"{map_(neg)}",
            marker=markers[a % len(markers)],
        )
    if frontier:
        ax.scatter(
            [frontier[0][0]],
            [frontier[0][1]],
            color="magenta",
            label=f"Max. Welfare",
            marker="s",
            alpha=WELFARE_ALPHA,
            s=int(default_marker_size * WELFARE_SCALE),
        )
        if show_annotations:
            ax.annotate(
                "Max. Welfare",
                xy=frontier[0],  # theta, radius
                xytext=(
                    frontier[0][0] + 0.02,
                    frontier[0][1] + 0.02 * yrange,
                ),  # fraction, fraction
                horizontalalignment="left",
                verticalalignment="bottom",
            )
    if nash:
        ax.scatter(
            [nash[0]],
            [nash[1]],
            color="cyan",
            label=f"Nash Point",
            marker="x",
            alpha=NASH_ALPHA,
            s=int(default_marker_size * NASH_SCALE),
        )
        if show_annotations:
            ax.annotate(
                "Nash Point",
                xy=nash,  # theta, radius
                xytext=(
                    nash[0] + 0.02,
                    nash[1] - 0.02 * yrange,
                ),  # fraction, fraction
                horizontalalignment="left",
                verticalalignment="bottom",
            )
    if agreement is not None:
        ax.scatter(
            [plotting_ufuns[0](agreement)],
            [plotting_ufuns[1](agreement)],
            color="black",
            marker="*",
            s=int(default_marker_size * AGREEMENT_SCALE),
            alpha=AGREEMENT_ALPHA,
            label="Agreement",
        )
        if show_annotations:
            ax.annotate(
                "Agreement",
                xy=nash,  # theta, radius
                xytext=(
                    agreement_utility[0] + 0.02,
                    agreement_utility[1] + 0.02,
                ),  # fraction, fraction
                horizontalalignment="left",
                verticalalignment="bottom",
            )
    ax.legend(
        bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
        loc="lower left",
        ncol=2,
        mode="expand",
        borderaxespad=0.0,
    )


def plot_mechanism_run(
    mechanism,
    negotiators: tuple[int, int] | tuple[str, str] | None = (0, 1),
    save_fig: bool = False,
    path: str | None = None,
    fig_name: str | None = None,
    ignore_none_offers: bool = True,
    with_lines: bool = True,
    show_agreement: bool = False,
    show_pareto_distance: bool = True,
    show_nash_distance: bool = True,
    show_end_reason: bool = True,
    show_last_negotiator: bool = True,
    show_annotations: bool = False,
    show_reserved: bool = True,
    colors: list | None = None,
    markers: list[str] | None = None,
    colormap: str = "jet",
    ylimits: tuple[float, float] | None = None,
    common_legend=True,
):
    import matplotlib.gridspec as gridspec
    import matplotlib.pyplot as plt

    if negotiators is None:
        negotiators = (0, 1)
    if len(negotiators) != 2:
        raise ValueError(
            f"Cannot plot the 2D plot for the mechanism run without knowing two plotting negotiators"
        )
    plotting_negotiators = []
    plotting_ufuns = []
    for n in negotiators:
        if isinstance(n, int):
            i = n
        else:
            i = mechanism.negotiator_index(n)
            if i is None:
                raise ValueError(f"Cannot find negotiator with ID {n}")
        plotting_negotiators.append(mechanism.negotiators[i].id)
        plotting_ufuns.append(mechanism.negotiators[i].ufun)

    fig = plt.figure()
    gs = gridspec.GridSpec(mechanism.nmi.n_negotiators, 2)
    axs = []
    colors, markers = make_colors_and_markers(
        colors, markers, len(mechanism.negotiators), colormap
    )

    name_map = dict(zip(mechanism.negotiator_ids, mechanism.negotiator_names))
    all_ufuns = [_.ufun for _ in mechanism.negotiators]
    for a, neg in enumerate(mechanism.negotiator_ids):
        if a == 0:
            axs.append(fig.add_subplot(gs[a, 1]))
        else:
            axs.append(fig.add_subplot(gs[a, 1], sharex=axs[0]))
        plot_offer_utilities(
            trace=mechanism.full_trace,
            negotiator=neg,
            plotting_ufuns=all_ufuns,
            plotting_negotiators=mechanism.negotiator_ids,
            ax=axs[-1],
            name_map=name_map,
            colors=colors,
            markers=markers,
            ignore_none_offers=ignore_none_offers,
            ylimits=ylimits,
            show_legend=not common_legend or a == 0,
            show_x_label=a == len(mechanism.negotiator_ids) - 1,
            show_reserved=show_reserved,
        )
    axu = fig.add_subplot(gs[:, 0])
    agreement = mechanism.agreement
    state = mechanism.state
    if not show_last_negotiator:
        last_negotiator = None
    else:
        last_negotiator = state.last_negotiator
    if not show_end_reason:
        reason = None
    else:
        if state.timedout:
            reason = "Negotiation Timedout"
        elif agreement is not None:
            reason = "Negotiation Success"
        elif state.has_error:
            reason = "Negotiation ERROR"
        elif agreement is not None:
            reason = "Agreemend Reached"
        elif state.broken:
            reason = "Negotiation Ended"
        elif agreement is None:
            reason = "No Agreement"
        else:
            reason = "Unknown state!!"

    plot_2dutils(
        trace=mechanism.full_trace,
        plotting_ufuns=plotting_ufuns,
        plotting_negotiators=plotting_negotiators,
        offering_negotiators=mechanism.negotiator_ids,
        outcome_space=mechanism.outcome_space,
        outcomes=mechanism.discrete_outcomes(),
        ax=axu,
        name_map=name_map,
        with_lines=with_lines,
        show_agreement=show_agreement,
        show_pareto_distance=show_pareto_distance,
        show_nash_distance=show_nash_distance,
        show_annotations=show_annotations,
        show_reserved=show_reserved,
        colors=colors,
        markers=markers,
        agreement=mechanism.state.agreement,
        end_reason=reason,
        last_negotiator=last_negotiator,
    )
    if save_fig:
        if fig_name is None:
            fig_name = str(uuid.uuid4()) + ".png"
        if path is None:
            path_ = pathlib.Path().absolute()
        else:
            path_ = pathlib.Path(path)
        path_.mkdir(parents=True, exist_ok=True)
        fig.savefig(str(path_ / fig_name), bbox_inches="tight", transparent=False)
    return fig



from __future__ import annotations

from .agent import Agent
from .entity import Entity

__all__ = ["Adapter"]


class Adapter(Agent):
    """
    Represents an adapter agent that makes some included object act as an
    agent in some world.

    Args:

        obj: The object to be adapted.
        include_adapter_type_name: Whether to include the adapter type name. If
                                   None, then it will be included if it does
                                   not start with and underscore.
        include_obj_type_name: Whether to include object type name in this
                               adapter's type name
        type_postfix: A string to add to the end of the type name

    Remarks:

        - Other than keeping an internal copy of the adapted object under
          `obj`, this class is used primarily to provide a way to give
          good type_name and short_type_name properties that combine the name
          of the adapter and the name of the enclosed object nicely.
        - The adapted object must be an `Entity`.
        - The `World` class uses the type names from this adapter whenever
          it needs to get a type-name (either `type_name` or `short_type_name`)
    """

    def __init__(
        self,
        obj,
        include_adapter_type_name: bool | None = None,
        include_obj_type_name=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if include_adapter_type_name is None:
            include_adapter_type_name = not self.__class__.__name__.startswith("_")

        self._obj = obj
        self._include_adapter, self._include_obj = (
            include_adapter_type_name,
            include_obj_type_name,
        )

    @property
    def adapted_object(self) -> Entity:
        return self._obj

    @adapted_object.setter
    def adapted_object(self, x: Entity) -> Entity:
        self._obj = x
        return x

    @property
    def short_type_name(self):
        """Returns a short name of the type of this entity"""
        base = super().short_type_name if self._include_adapter else ""
        obj = self._obj.short_type_name if self._include_obj else ""
        return base + ":" + obj

    @property
    def type_name(self):
        """Returns a short name of the type of this entity"""
        base = super().type_name if self._include_adapter else ""
        obj = self._obj.type_name if self._include_obj else ""
        return base + ":" + obj

    def init(self):
        """Override this method to modify initialization logic"""
        self._obj.init()

    def step(self):
        """Override this method to modify stepping logic"""
        self._obj.step()

    def __getattr__(self, attr):
        return getattr(self._obj, attr)



from __future__ import annotations

import random
from typing import Any, Callable, Iterable

from negmas.helpers import get_full_type_name
from negmas.helpers.numeric import get_one_int
from negmas.outcomes import Outcome
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, deserialize, serialize

from .base import Value
from .base_ufun import BaseUtilityFunction
from .crisp.linear import LinearAdditiveUtilityFunction

__all__ = ["WeightedUtilityFunction", "ComplexNonlinearUtilityFunction"]


class _DependenceMixin:
    """Used to set dependence properties based on the object's `values` ."""

    def is_session_dependent(self):  # type: ignore
        return any(_.is_session_dependent() for _ in self.values)  # type: ignore

    def is_stationary(self) -> bool:  # type: ignore
        return any(_.is_stationary() for _ in self.values)  # type: ignore

    def is_volatile(self) -> bool:  # type: ignore
        return any(_.is_volatile() for _ in self.values)  # type: ignore

    def is_state_dependent(self) -> bool:  # type: ignore
        return any(_.is_state_dependent() for _ in self.values)  # type: ignore


class WeightedUtilityFunction(_DependenceMixin, BaseUtilityFunction):
    """A utility function composed of linear aggregation of other utility functions

    Args:
        ufuns: An iterable of utility functions
        weights: Weights used for combination. If not given all weights are assumed to equal 1.
        name: Utility function name

    """

    def __init__(
        self,
        ufuns: Iterable[BaseUtilityFunction],
        weights: Iterable[float] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.values: list[BaseUtilityFunction] = list(ufuns)
        if weights is None:
            weights = [1.0] * len(self.values)
        self.weights = list(weights)

    def to_stationary(self):
        return WeightedUtilityFunction(
            ufuns=[_.to_stationary() for _ in self.values],
            weights=self.weights,
            name=self.name,
            id=self.id,
        )

    @classmethod
    def random(
        cls,
        outcome_space,
        reserved_value,
        normalized=True,
        n_ufuns=(1, 4),
        ufun_types=(LinearAdditiveUtilityFunction,),
        **kwargs,
    ) -> WeightedUtilityFunction:
        """Generates a random ufun of the given type"""
        n = get_one_int(n_ufuns)
        ufuns = [
            random.choice(ufun_types).random(outcome_space, 0, normalized)
            for _ in range(n)
        ]
        weights = [random.random() for _ in range(n)]
        return WeightedUtilityFunction(
            reserved_value=reserved_value,
            ufuns=ufuns,
            weights=weights,
            outcome_space=outcome_space,
            **kwargs,
        )

    def eval(self, offer: Outcome) -> Value:
        """Calculate the utility_function value for a given outcome.

        Args:
            offer: The offer to be evaluated.


        Remarks:
            - You cannot return None from overriden apply() functions but raise an exception (ValueError) if it was
              not possible to calculate the Value.
            - Return A Value not a float for real-valued utilities for the benefit of inspection code.

        Returns:
            Value: The utility_function value which may be a distribution. If `None` it means the utility_function value cannot be
            calculated.
        """
        if offer is None:
            return self.reserved_value
        u = float(0.0)
        for f, w in zip(self.values, self.weights):
            util = f(offer)
            if util is None or w is None:
                raise ValueError(
                    f"Cannot calculate utility for {offer}\n\t UFun {str(f)}\n\t with vars\n{vars(f)}"
                )
            u += util * w  # type: ignore
        return u

    def to_dict(self) -> dict[str, Any]:
        d = {PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self))}
        d.update(super().to_dict())
        return dict(
            **d,
            ufuns=[serialize(_) for _ in self.values],
            weights=self.weights,
        )

    @classmethod
    def from_dict(cls, d: dict[str, Any]):
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        d["ufuns"] = [deserialize(_) for _ in d["ufuns"]]
        return cls(**d)


class ComplexNonlinearUtilityFunction(_DependenceMixin, BaseUtilityFunction):
    """A utility function composed of nonlinear aggregation of other utility functions

    Args:
        ufuns: An iterable of utility functions
        combination_function: The function used to combine results of ufuns
        name: Utility function name

    """

    def __init__(
        self,
        ufuns: Iterable[BaseUtilityFunction],
        combination_function=Callable[[Iterable[Value]], Value],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.ufuns = list(ufuns)
        self.combination_function = combination_function

    def to_stationary(self):
        return ComplexNonlinearUtilityFunction(
            ufuns=[_.to_stationary() for _ in self.ufuns],
            combination_function=self.combination_function,
            name=self.name,
            id=self.id,
        )

    @classmethod
    def random(
        cls,
        outcome_space,
        reserved_value,
        normalized=True,
        n_ufuns=(1, 4),
        ufun_types=(LinearAdditiveUtilityFunction,),
        **kwargs,
    ) -> ComplexNonlinearUtilityFunction:
        """Generates a random ufun of the given type"""
        n = get_one_int(n_ufuns)
        ufuns = [
            random.choice(ufun_types).random(outcome_space, 0, normalized)
            for _ in range(n)
        ]
        weights = [random.random() for _ in range(n)]
        return ComplexNonlinearUtilityFunction(
            reserved_value=reserved_value,
            ufuns=ufuns,
            combination_function=lambda vals: sum(v * w for w, v in zip(weights, vals)),
            outcome_space=outcome_space,
            **kwargs,
        )

    def to_dict(self) -> dict[str, Any]:
        d = {PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self))}
        d.update(super().to_dict())
        return dict(
            ufuns=serialize(self.ufuns),
            combination_function=serialize(self.combination_function),
        )

    @classmethod
    def from_dict(cls, d: dict[str, Any]):
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        d["ufuns"] = deserialize(d["ufuns"])
        d["combination_function"] = deserialize(d["combination_function"])
        return cls(**d)

    def eval(self, offer: Outcome) -> Value:
        """Calculate the utility_function value for a given outcome.

        Args:
            offer: The offer to be evaluated.


        Remarks:
            - You cannot return None from overriden apply() functions but raise an exception (ValueError) if it was
              not possible to calculate the Value.
            - Return A Value not a float for real-valued utilities for the benefit of inspection code.

        Returns:
            Value: The utility_function value which may be a distribution. If `None` it means the utility_function value cannot be
            calculated.
        """
        if offer is None:
            return self.reserved_value
        return self.combination_function([f(offer) for f in self.ufuns])



from __future__ import annotations

import pytest

from negmas.genius import Atlas3, NiceTitForTat, genius_bridge_is_running
from negmas.outcomes import make_issue
from negmas.preferences import LinearUtilityFunction as U
from negmas.sao import AspirationNegotiator, NaiveTitForTatNegotiator
from negmas.situated.neg import NegDomain
from negmas.tournaments.neg import domains_from_list, neg_tournament


def test_can_run_world():
    issues = (make_issue(10, "quantity"), make_issue(5, "price"))
    competitors = [AspirationNegotiator, NaiveTitForTatNegotiator]
    if genius_bridge_is_running():
        competitors += [Atlas3, NiceTitForTat]

    domains = []
    for index in range(2):
        for partner in competitors:
            domains.append(
                NegDomain(
                    name="d0",
                    issues=issues,
                    ufuns=(
                        U.random(
                            issues=issues, reserved_value=(0.0, 0.2), normalized=False
                        ),
                        U.random(
                            issues=issues, reserved_value=(0.0, 0.2), normalized=False
                        ),
                    ),
                    partner_types=(partner,),
                    index=index,
                )
            )

    print(
        neg_tournament(
            n_configs=2 * 2,
            domains=domains_from_list(domains),
            competitors=competitors,
            n_steps=2,
            neg_n_steps=10,
            neg_time_limit=None,
            parallelism="serial",
        )
    )


@pytest.mark.skip(reason="no way of currently testing this")
def test_can_run_tournament():
    issues = (make_issue(10, "quantity"), make_issue(5, "price"))
    competitors = [AspirationNegotiator, NaiveTitForTatNegotiator]
    if genius_bridge_is_running():
        competitors += [Atlas3, NiceTitForTat]

    domains = []
    for index in range(2):
        for partner in competitors:
            domains.append(
                NegDomain(
                    name="d0",
                    issues=issues,
                    ufuns=(
                        U.random(issues, reserved_value=(0.0, 0.2), normalized=False),
                        U.random(issues, reserved_value=(0.0, 0.2), normalized=False),
                    ),
                    partner_types=(partner,),
                    index=index,
                )
            )

    neg_tournament(
        n_configs=len(domains),
        domains=domains_from_list(domains),
        competitors=competitors,
        n_steps=1,
        neg_n_steps=10,
        neg_time_limit=None,
        compact=True,
    )


@pytest.mark.skip(reason="no way of currently testing this")
def test_can_run_tournament_from_generator():
    from negmas.tournaments.neg import random_discrete_domains

    n_configs = 1
    n_repetitions = 1
    competitors = [AspirationNegotiator, NaiveTitForTatNegotiator]
    if genius_bridge_is_running():
        competitors += [Atlas3, NiceTitForTat]

    domains = random_discrete_domains(issues=[5, 4, (3, 5)], partners=competitors)

    neg_tournament(
        n_configs=len(competitors) * n_configs,
        domains=domains,
        competitors=competitors,
        n_steps=n_repetitions,
        neg_n_steps=10,
        neg_time_limit=None,
        compact=True,
    )



"""
Implements Stacked Alternating Offers (SAO) mechanism.
"""
from __future__ import annotations

import functools
import random
import sys
import time
from collections import defaultdict, namedtuple
from typing import TYPE_CHECKING

from negmas import warnings

from ..events import Event
from ..helpers import TimeoutCaller, TimeoutError, exception2str
from ..mechanisms import Mechanism, MechanismRoundResult
from ..outcomes.common import Outcome
from ..outcomes.outcome_ops import cast_value_types, outcome_types_are_ok
from .common import ResponseType, SAOResponse, SAOState

if TYPE_CHECKING:
    from negmas.preferences import Preferences

    from .negotiators import SAONegotiator

__all__ = [
    "SAOMechanism",
    "SAOProtocol",
    "TraceElement",
]

TraceElement = namedtuple(
    "TraceElement", ["time", "relative_time", "step", "negotiator", "offer", "response"]
)
"""An element of the trace returned by `full_trace` representing the history of the negotiation"""


class SAOMechanism(Mechanism):
    """
    Implements Several variants of the Stacked Alternating Offers Protocol

    Args:
        outcome_space: The negotiation agenda
        issues: A list of issues defining the outcome-space of the negotiation
        outcomes: A list of outcomes defining the outcome-space of the negotiation
        n_steps: The maximum number of negotiaion rounds (see `time_limit` )
        time_limit: The maximum wall-time allowed for the negotiation (see `n_steps`)
        step_time_limit: The maximum wall-time allowed for a single negotiation round.
        max_n_agents: The maximum number of negotiators allowed to join the negotiation.
        dynamic_entry: Whether it is allowed for negotiators to join the negotiation after it starts
        cache_outcomes: If true, the mechnism will catch `outcomes` and a discrete version (`discrete_outcomes`) that
                        can be accessed by any negotiator through their AMI.
        max_cardinality: Maximum number or outcomes to use when disctetizing the outcome-space
        annotation: A key-value mapping to keep around. Accessible through the AMI but not used by the mechanism.
        end_on_no_response: End the negotiation if any negotiator returns NO_RESPONSE from `respond`/`counter` or returns
                            REJECT_OFFER then refuses to give an offer (by returning `None` from `proposee/`counter`).
        enable_callbacks: Enable callbacks like on_round_start, etc. Note that on_negotiation_end is always received
                          by the negotiators no matter what is the setting for this parameter.
        avoid_ultimatum: If true, a proposal is taken from every agent in the first round then all of them are discarded
                         except one (choose randomly) and the algorithm continues using this one. This prevents negotiators
                         from knowing their order in the round.
        check_offers: If true, offers are checked to see if they are valid for the negotiation
                      outcome-space and if not the offer is considered None which is the same as
                      refusing to offer (NO_RESPONSE).
        enforce_issue_types: If True, the type of each issue is enforced depending on the value of `cast_offers`
        cast_offers: If true, each issue value is cast using the issue's type otherwise an incorrect type will be considered an invalid offer. See `check_offers`. Only
                     used if `enforce_issue_types`
        ignore_negotiator_exceptions: just silently ignore negotiator exceptions and consider them no-responses.
        offering_is_accepting: Offering an outcome implies accepting it. If not, the agent who proposed an offer will
                               be asked to respond to it after all other agents.
        name: Name of the mecnanism
        **kwargs: Extra paramters passed directly to the `Mechanism` constructor

    Remarks:

        - One and only one of `outcome_space`, `issues`, `outcomes` can be given
        - If both `n_steps` and `time_limit` are passed, the negotiation ends when either of the limits is reached.
        - Negotiations may take longer than `time_limit` because negotiators are not interrupted while they are
          executing their `respond` or `propose` methods.

    Events:

        - negotiator_exception: Data=(negotiator, exception) raised whenever a negotiator raises an exception if
          ignore_negotiator_exceptions is set to True.
    """

    def __init__(
        self,
        dynamic_entry=False,
        extra_callbacks=False,
        end_on_no_response=True,
        avoid_ultimatum=False,
        check_offers=False,
        enforce_issue_types=False,
        cast_offers=False,
        ignore_negotiator_exceptions=False,
        offering_is_accepting=True,
        allow_offering_just_rejected_outcome=True,
        name: str | None = None,
        max_wait: int = sys.maxsize,
        sync_calls: bool = False,
        **kwargs,
    ):
        kwargs["state_factory"] = SAOState
        super().__init__(
            dynamic_entry=dynamic_entry,
            extra_callbacks=extra_callbacks,
            name=name,
            **kwargs,
        )
        self._current_state: SAOState
        n_steps, time_limit = self.n_steps, self.time_limit
        if (n_steps is None or n_steps == float("inf")) and (
            time_limit is None or time_limit == float("inf")
        ):
            warnings.warn(
                "You are passing no time_limit and no n_steps to an SAOMechanism. The mechanism may never finish!!",
                warnings.NegmasInfiniteNegotiationWarning,
            )
        self._sync_calls = sync_calls
        self.params["end_on_no_response"] = end_on_no_response
        self.params["enable_callbacks"] = extra_callbacks
        self.params["avoid_ultimatum"] = avoid_ultimatum
        self.params["check_offers"] = check_offers
        self.params["offering_is_accepting"] = offering_is_accepting
        self.params["enforce_issue_types"] = enforce_issue_types
        self.params["cast_offers"] = cast_offers
        self.params[
            "allow_offering_just_rejected_outcome"
        ] = allow_offering_just_rejected_outcome
        self._n_max_waits = max_wait if max_wait is not None else float("inf")
        self.params["max_wait"] = self._n_max_waits
        self.ignore_negotiator_exceptions = ignore_negotiator_exceptions
        self.allow_offering_just_rejected_outcome = allow_offering_just_rejected_outcome
        self.end_negotiation_on_refusal_to_propose = end_on_no_response
        self.check_offers = check_offers
        self._enforce_issue_types = enforce_issue_types
        self._cast_offers = cast_offers

        self._last_checked_negotiator = -1
        self._current_proposer = None
        self._frozen_neg_list = None
        self._avoid_ultimatum = n_steps is not None and avoid_ultimatum
        self._ultimatum_avoided = False
        self._no_responses = 0
        self._offering_is_accepting = offering_is_accepting
        self._n_waits = 0
        self._waiting_time: dict[str, float] = defaultdict(float)
        self._waiting_start: dict[str, float] = defaultdict(lambda: float("inf"))
        self._selected_first = 0

    def add(
        self,
        negotiator: SAONegotiator,
        *,
        preferences: Preferences | None = None,
        role: str | None = None,
        **kwargs,
    ) -> bool | None:
        from ..genius.negotiator import GeniusNegotiator

        added = super().add(negotiator, preferences=preferences, role=role, **kwargs)
        if (
            added
            and isinstance(negotiator, GeniusNegotiator)
            and self.nmi.time_limit is not None
            and self.nmi.time_limit != float("inf")
            and self.nmi.n_steps is not None
        ):
            warnings.warn(
                f"{negotiator.id} of type {negotiator.__class__.__name__} is joining SAOMechanism which has a time_limit of {self.nmi.time_limit} seconds and a n_steps of {self.nmi.n_steps}. This agnet will only know about the time_limit and will not know about the n_steps!!!",
                warnings.NegmasStepAndTimeLimitWarning,
            )
        return added

    def set_sync_call(self, v: bool):
        self._sync_call = v

    def _agent_info(self):
        state = self._current_state
        current_proposer_agent = (
            self._current_proposer.owner if self._current_proposer else None
        )
        if current_proposer_agent:
            current_proposer_agent = current_proposer_agent.id
        new_offerer_agents = []
        for neg_id, _ in state.new_offers:
            neg = self._negotiator_map.get(neg_id, None)
            agent = neg.owner if neg else None
            if agent is not None:
                new_offerer_agents.append(agent.id)
            else:
                new_offerer_agents.append(None)
        return current_proposer_agent, new_offerer_agents

    def _stop_waiting(self, negotiator_id):
        self._waiting_time[negotiator_id] = 0.0
        self._waiting_start[negotiator_id] = float("inf")
        self._n_waits = 0
        self._frozen_neg_list = None

    def round(self) -> MechanismRoundResult:
        """implements a round of the Stacked Alternating Offers Protocol."""
        state = self._current_state
        if self._frozen_neg_list is None:
            state.new_offers = []
        negotiators: list[SAONegotiator] = self.negotiators
        n_negotiators = len(negotiators)
        # times = dict(zip([_.id for _ in negotiators], itertools.repeat(0.0)))
        times = defaultdict(float, self._waiting_time)
        exceptions = dict(
            zip([_.id for _ in negotiators], [list() for _ in negotiators])
        )

        def _safe_counter(
            negotiator, *args, **kwargs
        ) -> tuple[SAOResponse | None, bool]:
            assert (
                not state.waiting or negotiator.id == state.current_proposer
            ), f"We are waiting with {state.current_proposer} as the last offerer but we are asking {negotiator.id} to offer\n{state}"
            rem = self.remaining_time
            if rem is None:
                rem = float("inf")
            timeout = min(
                self.nmi.negotiator_time_limit - times[negotiator.id],
                self.nmi.step_time_limit,
                rem,
                self._hidden_time_limit - self.time,
            )
            if timeout is None or timeout == float("inf") or self._sync_calls:
                __strt = time.perf_counter()
                try:
                    if (
                        negotiator == self._current_proposer
                    ) and self._offering_is_accepting:
                        state.n_acceptances = 0
                        response = negotiator.counter(*args, **kwargs)
                    else:
                        response = negotiator.counter(*args, **kwargs)
                except TimeoutError:
                    response = None
                    try:
                        negotiator.cancel()
                    except:
                        pass
                except Exception as ex:
                    exceptions[negotiator.id].append(exception2str())
                    if self.ignore_negotiator_exceptions:
                        self.announce(
                            Event(
                                "negotiator_exception",
                                {"negotiator": negotiator, "exception": ex},
                            )
                        )
                        times[negotiator.id] += time.perf_counter() - __strt
                        return SAOResponse(ResponseType.END_NEGOTIATION, None), True
                    else:
                        raise ex
                times[negotiator.id] += time.perf_counter() - __strt
            else:
                fun = functools.partial(negotiator.counter, *args, **kwargs)
                __strt = time.perf_counter()
                try:
                    if (
                        negotiator == self._current_proposer
                    ) and self._offering_is_accepting:
                        state.n_acceptances = 0
                        response = TimeoutCaller.run(fun, timeout=timeout)
                    else:
                        response = TimeoutCaller.run(fun, timeout=timeout)
                except TimeoutError:
                    response = None
                except Exception as ex:
                    exceptions[negotiator.id].append(exception2str())
                    if self.ignore_negotiator_exceptions:
                        self.announce(
                            Event(
                                "negotiator_exception",
                                {"negotiator": negotiator, "exception": ex},
                            )
                        )
                        times[negotiator.id] += time.perf_counter() - __strt
                        return SAOResponse(ResponseType.END_NEGOTIATION, None), True
                    else:
                        raise ex
                times[negotiator.id] += time.perf_counter() - __strt
            if (
                self.check_offers
                and response is not None
                and response.outcome is not None
            ):
                if not self.outcome_space.is_valid(response.outcome):
                    return SAOResponse(response.response, None), False
                # todo: do not use .issues here as they are not guaranteed to exist (if it is not a cartesial outcome space)
                if self._enforce_issue_types and hasattr(self.outcome_space, "issues"):
                    if outcome_types_are_ok(
                        response.outcome, self.outcome_space.issues  # type: ignore
                    ):
                        return response, False
                    elif self._cast_offers:
                        return (
                            SAOResponse(
                                response.response,
                                cast_value_types(
                                    response.outcome, self.outcome_space.issues  # type: ignore
                                ),
                            ),
                            False,
                        )
                    return SAOResponse(response.response, None), False
            return response, False

        proposers, proposer_indices = [], []
        for i, neg in enumerate(negotiators):
            if not neg.capabilities.get("propose", False):
                continue
            proposers.append(neg)
            proposer_indices.append(i)
        n_proposers = len(proposers)
        if n_proposers < 1:
            if not self.dynamic_entry:
                return MechanismRoundResult(
                    broken=True,
                    timedout=False,
                    agreement=None,
                    error=True,
                    error_details="No proposers and no dynamic entry",
                    times=times,
                    exceptions=exceptions,
                )
            else:
                return MechanismRoundResult(
                    broken=False,
                    timedout=False,
                    agreement=None,
                    times=times,
                    exceptions=exceptions,
                )
        # if this is the first step (or no one has offered yet) which means that there is no _current_offer
        if (
            state.current_offer is None
            and n_proposers > 1
            and self._avoid_ultimatum
            and not self._ultimatum_avoided
        ):
            if not self.dynamic_entry and not self.state.step == 0:
                if self.end_negotiation_on_refusal_to_propose:
                    return MechanismRoundResult(
                        broken=True,
                        times=times,
                        exceptions=exceptions,
                    )
            # if we are trying to avoid an ultimatum, we take an offer from everyone and ignore them but one.
            # this way, the agent cannot know its order. For example, if we have two agents and 3 steps, this will
            # be the situation after each step:
            #
            # Case 1: Assume that it ignored the offer from agent 1
            # Step, Agent 0 calls received  , Agent 1 calls received    , relative time during last call
            # 0   , counter(None)->offer1*  , counter(None) -> offer0   , 0/3
            # 1   , counter(offer2)->offer3 , counter(offer1) -> offer2 , 1/3
            # 2   , counter(offer4)->offer5 , counter(offer3) -> offer4 , 2/3
            # 3   ,                         , counter(offer5)->offer6   , 3/3
            #
            # Case 2: Assume that it ignored the offer from agent 0
            # Step, Agent 0 calls received  , Agent 1 calls received    , relative time during last call
            # 0   , counter(None)->offer1   , counter(None) -> offer0*  , 0/3
            # 1   , counter(offer0)->offer2 , counter(offer2) -> offer3 , 1/3
            # 2   , counter(offer3)->offer4 , counter(offer4) -> offer5 , 2/3
            # 3   , counter(offer5)->offer6 ,                           , 3/3
            #
            # in both cases, the agent cannot know whether its last offer going to be passed to the other agent
            # (the ultimatum scenario) or not.
            responses = []
            responders = []
            for i, neg in enumerate(proposers):
                if not neg.capabilities.get("propose", False):
                    continue
                strt = time.perf_counter()
                resp, has_exceptions = _safe_counter(neg, state=self.state, offer=None)
                if has_exceptions:
                    return MechanismRoundResult(
                        broken=True,
                        timedout=False,
                        agreement=None,
                        times=times,
                        exceptions=exceptions,
                        error=True,
                        error_details=str(exceptions[neg.id]),
                    )
                if resp is None:
                    return MechanismRoundResult(
                        broken=False,
                        timedout=True,
                        agreement=None,
                        times=times,
                        exceptions=exceptions,
                        error=False,
                        error_details="",
                    )
                if resp.response != ResponseType.WAIT:
                    self._waiting_time[neg.id] = 0.0
                    self._waiting_start[neg.id] = float("inf")
                    self._frozen_neg_list = None
                else:
                    self._waiting_start[neg.id] = min(self._waiting_start[neg.id], strt)
                    self._waiting_time[neg.id] += (
                        time.perf_counter() - self._waiting_start[neg.id]
                    )
                if resp is None:
                    return MechanismRoundResult(
                        broken=False,
                        timedout=True,
                        agreement=None,
                        times=times,
                        exceptions=exceptions,
                    )
                if time.perf_counter() - strt > self.nmi.step_time_limit:
                    return MechanismRoundResult(
                        broken=False,
                        timedout=True,
                        agreement=None,
                        times=times,
                        exceptions=exceptions,
                    )
                if resp.response == ResponseType.END_NEGOTIATION:
                    return MechanismRoundResult(
                        broken=True,
                        timedout=False,
                        agreement=None,
                        times=times,
                        exceptions=exceptions,
                    )
                if resp.response in (ResponseType.NO_RESPONSE, ResponseType.WAIT):
                    continue
                if (
                    resp.response == ResponseType.REJECT_OFFER
                    and resp.outcome is None
                    and self.end_negotiation_on_refusal_to_propose
                ):
                    continue
                responses.append(resp)
                responders.append(i)
            if len(responses) < 1:
                if not self.dynamic_entry:
                    return MechanismRoundResult(
                        broken=True,
                        timedout=False,
                        agreement=None,
                        error=True,
                        error_details="No proposers and no dynamic entry. This may happen if no negotiators responded to their first proposal request with an offer",
                        times=times,
                        exceptions=exceptions,
                    )
                else:
                    return MechanismRoundResult(
                        broken=False,
                        timedout=False,
                        agreement=None,
                        times=times,
                        exceptions=exceptions,
                    )
            # choose a random negotiator and set it as the current negotiator
            self._ultimatum_avoided = True
            selected = random.randint(0, len(responses) - 1)
            resp = responses[selected]
            neg = proposers[responders[selected]]
            _first_proposer = proposer_indices[responders[selected]]
            self._selected_first = _first_proposer
            self._last_checked_negotiator = _first_proposer
            state.current_offer = resp.outcome
            state.new_offers.append((neg.id, resp.outcome))
            self._current_proposer = neg
            state.current_proposer = neg.id
            state.n_acceptances = 1 if self._offering_is_accepting else 0
            if self._last_checked_negotiator >= 0:
                state.last_negotiator = self.negotiators[
                    self._last_checked_negotiator
                ].name
            else:
                state.last_negotiator = ""
            (
                self._current_proposer_agent,
                state.new_offerer_agents,
            ) = self._agent_info()

            # current_proposer_agent=current_proposer_agent,
            # new_offerer_agents=new_offerer_agents,
            return MechanismRoundResult(
                broken=False,
                timedout=False,
                agreement=None,
                times=times,
                exceptions=exceptions,
            )

        # this is not the first round. A round will get n_negotiators responses
        if self._frozen_neg_list is not None:
            ordered_indices = self._frozen_neg_list
        else:
            ordered_indices = [
                (_ + self._last_checked_negotiator + 1) % n_negotiators
                for _ in range(n_negotiators)
            ]

        for _, neg_indx in enumerate(ordered_indices):
            self._last_checked_negotiator = neg_indx
            neg = self.negotiators[neg_indx]
            strt = time.perf_counter()
            resp, has_exceptions = _safe_counter(
                neg, state=self.state, offer=state.current_offer
            )
            if has_exceptions:
                return MechanismRoundResult(
                    broken=True,
                    timedout=False,
                    agreement=None,
                    times=times,
                    exceptions=exceptions,
                    error=True,
                    error_details=str(exceptions[neg.id]),
                )
            if resp is None:
                return MechanismRoundResult(
                    broken=False,
                    timedout=True,
                    agreement=None,
                    times=times,
                    exceptions=exceptions,
                    error=False,
                    error_details="",
                )
            if resp.response == ResponseType.WAIT:
                self._waiting_start[neg.id] = min(self._waiting_start[neg.id], strt)
                self._waiting_time[neg.id] += time.perf_counter() - strt
                self._last_checked_negotiator = (neg_indx - 1) % n_negotiators
                offered = {self._negotiator_index[_[0]] for _ in state.new_offers}
                did_not_offer = sorted(
                    list(set(range(n_negotiators)).difference(offered))
                )
                assert neg_indx in did_not_offer
                indx = did_not_offer.index(neg_indx)
                assert (
                    self._frozen_neg_list is None
                    or self._frozen_neg_list[0] == neg_indx
                )
                self._frozen_neg_list = did_not_offer[indx:] + did_not_offer[:indx]
                self._n_waits += 1
            else:
                self._stop_waiting(neg.id)

            if resp is None:
                return MechanismRoundResult(
                    broken=False,
                    timedout=True,
                    agreement=None,
                    times=times,
                    exceptions=exceptions,
                )
            if time.perf_counter() - strt > self.nmi.step_time_limit:
                return MechanismRoundResult(
                    broken=False,
                    timedout=True,
                    agreement=None,
                    times=times,
                    exceptions=exceptions,
                )
            if self._extra_callbacks:
                if state.current_offer is not None:
                    for other in self.negotiators:
                        if other is not neg:
                            other.on_partner_response(
                                state=self.state,
                                partner_id=neg.id,
                                outcome=state.current_offer,
                                response=resp.response,
                            )
            if resp.response == ResponseType.NO_RESPONSE:
                continue
            if resp.response == ResponseType.WAIT:
                if self._n_waits > self._n_max_waits:
                    self._stop_waiting(neg.id)
                    return MechanismRoundResult(
                        broken=False,
                        timedout=True,
                        agreement=None,
                        waiting=False,
                        times=times,
                        exceptions=exceptions,
                    )
                return MechanismRoundResult(
                    broken=False,
                    timedout=False,
                    agreement=None,
                    waiting=True,
                    times=times,
                    exceptions=exceptions,
                )
            if resp.response == ResponseType.END_NEGOTIATION:
                return MechanismRoundResult(
                    broken=True,
                    timedout=False,
                    agreement=None,
                    times=times,
                    exceptions=exceptions,
                )
            if resp.response == ResponseType.ACCEPT_OFFER:
                state.n_acceptances += 1
                if state.n_acceptances == n_negotiators:
                    return MechanismRoundResult(
                        broken=False,
                        timedout=False,
                        agreement=state.current_offer,
                        times=times,
                        exceptions=exceptions,
                    )
            if resp.response == ResponseType.REJECT_OFFER:
                proposal = resp.outcome
                if (
                    not self.allow_offering_just_rejected_outcome
                    and proposal == state.current_offer
                ):
                    proposal = None
                if proposal is None:
                    if (
                        neg.capabilities.get("propose", True)
                        and self.end_negotiation_on_refusal_to_propose
                    ):
                        return MechanismRoundResult(
                            broken=True,
                            timedout=False,
                            agreement=None,
                            times=times,
                            exceptions=exceptions,
                        )
                    state.n_acceptances = 0
                else:
                    state.n_acceptances = 1 if self._offering_is_accepting else 0
                    if self._extra_callbacks:
                        for other in self.negotiators:
                            if other is neg:
                                continue
                            other.on_partner_proposal(
                                partner_id=neg.id, offer=proposal, state=self.state
                            )
                state.current_offer = proposal
                self._current_proposer = neg
                state.current_proposer = neg.id
                state.new_offers.append((neg.id, proposal))
                if self._last_checked_negotiator >= 0:
                    state.last_negotiator = self.negotiators[
                        self._last_checked_negotiator
                    ].name
                else:
                    state.last_negotiator = ""
                (
                    self._current_proposer_agent,
                    state.new_offerer_agents,
                ) = self._agent_info()

        return MechanismRoundResult(
            broken=False,
            timedout=False,
            agreement=None,
            times=times,
            exceptions=exceptions,
        )

    @property
    def full_trace(self) -> list[TraceElement]:
        """Returns the negotiation history as a list of relative_time/step/negotiator/offer tuples"""

        def response(state: SAOState):
            if state.timedout:
                return "timedout"
            if state.ended:
                return "ended"
            if state.has_error:
                return "error"
            if state.agreement:
                return "accepted"
            return "rejected"

        offers = []
        for state in self._history:
            state: SAOState
            offers += [
                TraceElement(
                    state.time, state.relative_time, state.step, n, o, response(state)
                )
                for n, o in state.new_offers
            ]

        def not_equal(a, b):
            return any(x != y for x, y in zip(a, b))

        self._history: list[SAOState]
        # if the agreement does not appear as the last offer in the trace, add it.
        # this should not happen though!!
        if (
            self.agreement is not None
            and offers
            and not_equal(offers[-1][-1], self.agreement)
        ):
            offers.append(
                TraceElement(
                    self._history[-1].time,
                    self._history[-1].relative_time,
                    self._history[-1].step,
                    self._history[-1].current_proposer,
                    self.agreement,
                    response(self._history[-1]),
                )
            )

        return offers

    @property
    def extended_trace(self) -> list[tuple[int, str, Outcome]]:
        """Returns the negotiation history as a list of step/negotiator/offer tuples"""
        offers = []
        for state in self._history:
            state: SAOState
            offers += [(state.step, n, o) for n, o in state.new_offers]

        def not_equal(a, b):
            return any(x != y for x, y in zip(a, b))

        self._history: list[SAOState]
        # if the agreement does not appear as the last offer in the trace, add it.
        # this should not happen though!!
        if (
            self.agreement is not None
            and offers
            and not_equal(offers[-1][-1], self.agreement)
        ):
            offers.append(
                (
                    self._history[-1].step,
                    self._history[-1].current_proposer,
                    self.agreement,
                )
            )

        return offers

    @property
    def trace(self) -> list[tuple[str, Outcome]]:
        """Returns the negotiation history as a list of negotiator/offer tuples"""
        offers = []
        for state in self._history:
            offers += [(n, o) for n, o in state.new_offers]

        def not_equal(a, b):
            if isinstance(a, dict):
                a = a.values()
            if isinstance(b, dict):
                b = b.values()
            return any(x != y for x, y in zip(a, b))

        if (
            self.agreement is not None
            and offers
            and not_equal(offers[-1][-1], self.agreement)
        ):
            offers.append(
                (
                    self._history[-1].current_proposer,
                    self.agreement,
                )
            )

        return offers

    def negotiator_offers(self, negotiator_id: str) -> list[Outcome]:
        """Returns the offers given by a negotiator (in order)"""
        return [o for n, o in self.trace if n == negotiator_id]

    def negotiator_full_trace(
        self, negotiator_id: str
    ) -> list[tuple[float, float, int, Outcome, str]]:
        """Returns the (time/relative-time/step/outcome/response) given by a negotiator (in order)"""
        return [
            (t, rt, s, o, r)
            for t, rt, s, n, o, r in self.full_trace
            if n == negotiator_id
        ]

    @property
    def offers(self) -> list[Outcome]:
        """Returns the negotiation history as a list of offers"""
        return [o for _, o in self.trace]

    @property
    def _step(self):
        """
        A private property used by the checkpoint system
        """
        return self._current_state.step

    def plot(
        self,
        plotting_negotiators: tuple[int, int] | tuple[str, str] = (0, 1),
        save_fig: bool = False,
        path: str | None = None,
        fig_name: str | None = None,
        ignore_none_offers: bool = True,
        with_lines: bool = True,
        show_agreement: bool = False,
        show_pareto_distance: bool = True,
        show_nash_distance: bool = True,
        show_end_reason: bool = True,
        show_last_negotiator: bool = True,
        show_annotations: bool = False,
        show_reserved: bool = True,
        colors: list | None = None,
        markers: list[str] | None = None,
        colormap: str = "jet",
        ylimits: tuple[float, float] | None = None,
    ):
        from negmas.sao.plots import plot_mechanism_run

        return plot_mechanism_run(
            self,
            negotiators=plotting_negotiators,
            save_fig=save_fig,
            path=path,
            fig_name=fig_name,
            ignore_none_offers=ignore_none_offers,
            with_lines=with_lines,
            show_agreement=show_agreement,
            show_pareto_distance=show_pareto_distance,
            show_nash_distance=show_nash_distance,
            show_end_reason=show_end_reason,
            show_last_negotiator=show_last_negotiator,
            show_annotations=show_annotations,
            show_reserved=show_reserved,
            colors=colors,
            markers=markers,
            colormap=colormap,
            ylimits=ylimits,
        )


SAOProtocol = SAOMechanism
"""An alias for `SAOMechanism` object"""



"""
Implements single text negotiation mechanisms
"""
from __future__ import annotations

import time
from copy import deepcopy

from attr import define, field

from .mechanisms import Mechanism, MechanismRoundResult, MechanismState
from .outcomes import Outcome

__all__ = [
    "VetoMTMechanism",
]


@define
class MTState(MechanismState):
    """Defines extra values to keep in the mechanism state. This is accessible to all negotiators"""

    current_offers: list[Outcome | None] = field(factory=list)


class VetoMTMechanism(Mechanism):
    """Base class for all muti-text mechanisms

    Args:
        *args: positional arguments to be passed to the base Mechanism
        **kwargs: keyword arguments to be passed to the base Mechanism
        n_texts: Number of current offers to keep
        initial_outcomes: initial set of outcomes outcome. If None, it will be selected by `next_outcome` which by default will choose it
                         randomly.
        initial_responses: Initial set of responses. One list per outcome in initial_outcomes.

    Remarks:

        - initial_responses is only of value when the number of negotiators that will join the negotiation is less then
          or equal to its length. By default it is not used for anything. Nevertheless, it is here because
          `next_outcome` may decide to use it with the `initial_outcome`


    """

    def __init__(
        self,
        *args,
        epsilon: float = 1e-6,
        n_texts: int = 10,
        initial_outcomes: list[Outcome | None] | None = None,
        initial_responses: tuple[tuple[bool]] | None = None,
        **kwargs,
    ):
        kwargs["state_factory"] = MTState
        super().__init__(*args, **kwargs)
        self._current_state: MTState
        state = self._current_state

        self.add_requirements(
            {"compare-binary": True}
        )  # assert that all agents must have compare-binary capability
        state.current_offers = (
            initial_outcomes if initial_outcomes is not None else [None] * n_texts
        )
        """The current offer"""
        self.initial_outcomes = deepcopy(state.current_offers)
        """The initial offer"""
        self.last_responses = (
            [list(_) for _ in initial_responses]
            if initial_responses is not None
            else [None] * n_texts
        )
        """The responses of all negotiators for the last offer"""
        self.initial_responses = deepcopy(self.last_responses)
        """The initial set of responses. See the remarks of this class to understand its role."""
        self.epsilon = epsilon

    def next_outcome(self, outcome: Outcome | None) -> Outcome | None:
        """Generate the next outcome given some outcome.

        Args:
             outcome: The current outcome

        Returns:
            a new outcome or None to end the mechanism run

        """
        return self.random_outcomes(1)[0]

    def round(self) -> MechanismRoundResult:
        """Single round of the protocol"""
        state: SAOState = self._current_state  # type: ignore
        for i, current_offer in enumerate(state.current_offers):
            new_offer = self.next_outcome(current_offer)
            responses = []

            for neg in self.negotiators:
                strt = time.perf_counter()
                responses.append(
                    neg.is_better(new_offer, current_offer, epsilon=self.epsilon)
                    is not False
                )
                if time.perf_counter() - strt > self.nmi.step_time_limit:
                    return MechanismRoundResult(
                        broken=False, timedout=True, agreement=None
                    )

            self.last_responses = responses

            if all(responses):
                state.current_offers[i] = new_offer

        return MechanismRoundResult(broken=False, timedout=False, agreement=None)

    def on_negotiation_end(self) -> None:
        """Used to pass the final offer for agreement between all negotiators"""
        self._agreement = self._current_state.current_offers

        super().on_negotiation_end()



from __future__ import annotations

from negmas.outcomes.cardinal_issue import CardinalIssue

__all__ = ["RangeIssue"]


class RangeIssue(CardinalIssue):
    """
    An issue representing a range of values (can be continuous or discrete)
    """

    def __init__(self, values, name=None) -> None:
        if values[1] < values[0]:
            raise ValueError(
                f"Trying to create a range issue with min {values[0]} and max {values[1]}!!!"
            )
        super().__init__(values, name)
        self._value_type = type(values[0])
        self.min_value, self.max_value = values[0], values[1]

    def is_valid(self, v):
        return self.min_value <= v <= self.max_value



from __future__ import annotations

import pathlib
from math import isclose

import numpy as np
import pkg_resources
import pytest
from numpy.testing import assert_almost_equal

from negmas import (
    GeniusNegotiator,
    genius_bridge_is_running,
    load_genius_domain_from_folder,
)
from negmas.genius import AgentX, Atlas3, Caduceus, GeniusBridge, YXAgent
from negmas.genius.gnegotiators import AgentK

DOMAIN_FOLDER = pathlib.Path(
    pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/scenarios/anac/y2010/Travel"
    )
)
DOMAIN_FILE = DOMAIN_FOLDER / "travel_domain.xml"
UTIL1 = DOMAIN_FOLDER / "travel_chox.xml"
UTIL2 = DOMAIN_FOLDER / "travel_fanny.xml"


SKIP_IF_NO_BRIDGE = True


@pytest.fixture(scope="module")
def init_genius():
    GeniusBridge.start(0)


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_agent(init_genius):
    domain = load_genius_domain_from_folder(DOMAIN_FOLDER)
    p = domain.make_session(
        [
            GeniusNegotiator(
                java_class_name="agents.anac.y2015.Atlas3.Atlas3",
                domain_file_name=DOMAIN_FILE,
                utility_file_name=UTIL1,
            ),
            GeniusNegotiator(
                java_class_name="agents.anac.y2015.AgentX.AgentX",
                domain_file_name=DOMAIN_FILE,
                utility_file_name=UTIL2,
            ),
        ],
        time_limit=120,
        n_steps=None,
    )
    issues = domain.issues
    assert len(p.negotiators) > 1
    issue_list = [f"{k}:{v}" for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)'"
        ", 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo', 'Festivals'"
        ", 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.', 'Monuments'"
        ", 'Special streets', 'Palace', 'Landscape and nature']",
    ]
    first, second = p.negotiators[0], p.negotiators[1]
    p.run()
    # print(f'{len(p.history)} bids exchanged')
    u1 = np.array([float(first.ufun(s.current_offer)) for s in p.history])
    u2 = np.array([float(second.ufun(s.current_offer)) for s in p.history])
    u1 + u2
    # print(
    #     f'Negotiator 1: {u1.mean()}({u1.std()})[{u1.min()}, {u1.max()}]\nNegotiator 2: {u2.mean()}({u2.std()})[{u1.min()}, {u1.max()}]'
    #     f'\nWelfare: {welfare.mean()}({welfare.std()})[{welfare.min()}, {welfare.max()}]')
    # print(p.state)
    assert len(u1) > 0


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_agent_top2016_caduceus_first(init_genius):
    domain = load_genius_domain_from_folder(DOMAIN_FOLDER)
    issues = domain.issues
    p = domain.make_session(
        [
            Caduceus(
                domain_file_name=DOMAIN_FILE,
                utility_file_name=UTIL1,
            ),
            AgentX(
                domain_file_name=DOMAIN_FILE,
                utility_file_name=UTIL2,
            ),
        ],
        time_limit=20,
        n_steps=None,
    )
    assert len(p.negotiators) > 1
    issue_list = [f"{k}:{v}" for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)'"
        ", 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo', 'Festivals'"
        ", 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.', 'Monuments'"
        ", 'Special streets', 'Palace', 'Landscape and nature']",
    ]
    first, second = p.negotiators[0], p.negotiators[1]
    p.run()
    # print(f'{len(p.history)} bids exchanged')
    u1 = np.array([float(first.ufun(s.current_offer)) for s in p.history])
    u2 = np.array([float(second.ufun(s.current_offer)) for s in p.history])
    u1 + u2
    # print(
    #     f'Negotiator 1: {u1.mean()}({u1.std()})[{u1.min()}, {u1.max()}]\nNegotiator 2: {u2.mean()}({u2.std()})[{u1.min()}, {u1.max()}]'
    #     f'\nWelfare: {welfare.mean()}({welfare.std()})[{welfare.min()}, {welfare.max()}]')
    # print(p.state)
    assert len(u1) > 0


def do_run(first_type, second_type):
    domain = load_genius_domain_from_folder(DOMAIN_FOLDER)
    issues = domain.issues
    p = domain.make_session(
        [
            first_type(domain_file_name=DOMAIN_FILE, utility_file_name=UTIL1),
            second_type(domain_file_name=DOMAIN_FILE, utility_file_name=UTIL2),
        ],
        time_limit=20,
        n_steps=None,
    )
    assert len(p.negotiators) > 1
    issue_list = [f"{k}:{v}" for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)'"
        ", 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo', 'Festivals'"
        ", 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.', 'Monuments'"
        ", 'Special streets', 'Palace', 'Landscape and nature']",
    ]
    first, second = p.negotiators[0], p.negotiators[1]
    p.run()
    # print(f'{len(p.history)} bids exchanged')
    u1 = np.array([float(first.ufun(s.current_offer)) for s in p.history])
    u2 = np.array([float(second.ufun(s.current_offer)) for s in p.history])
    # welfare = u1 + u2
    # print(
    #     f'Negotiator 1: {u1.mean()}({u1.std()})[{u1.min()}, {u1.max()}]\nNegotiator 2: {u2.mean()}({u2.std()})[{u1.min()}, {u1.max()}]'
    #     f'\nWelfare: {welfare.mean()}({welfare.std()})[{welfare.min()}, {welfare.max()}]')
    # print(p.state)
    assert len(u1) > 0 or len(u2) > 0


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
@pytest.mark.parametrize("first_type, second_type", [(AgentX, YXAgent)])
def test_genius_agent_top2016_yx_second_classes(init_genius, first_type, second_type):
    do_run(first_type, second_type)


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_agent_top2016_yx_second(init_genius):
    domain = load_genius_domain_from_folder(DOMAIN_FOLDER)
    issues = domain.issues
    p = domain.make_session(
        [
            GeniusNegotiator(
                java_class_name="agents.anac.y2016.yxagent.YXAgent",
                domain_file_name=DOMAIN_FILE,
                utility_file_name=UTIL1,
            ),
            GeniusNegotiator(
                java_class_name="agents.anac.y2015.AgentX.AgentX",
                domain_file_name=DOMAIN_FILE,
                utility_file_name=UTIL2,
            ),
        ],
        time_limit=20,
        n_steps=None,
    )
    assert len(p.negotiators) > 1
    issue_list = [f"{k}:{v}" for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)'"
        ", 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo', 'Festivals'"
        ", 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.', 'Monuments'"
        ", 'Special streets', 'Palace', 'Landscape and nature']",
    ]
    first, second = p.negotiators[0], p.negotiators[1]
    p.run()
    # print(f'{len(p.history)} bids exchanged')
    u1 = np.array([float(first.ufun(s.current_offer)) for s in p.history])
    u2 = np.array([float(second.ufun(s.current_offer)) for s in p.history])
    u1 + u2
    # print(
    #     f'Negotiator 1: {u1.mean()}({u1.std()})[{u1.min()}, {u1.max()}]\nNegotiator 2: {u2.mean()}({u2.std()})[{u1.min()}, {u1.max()}]'
    #     f'\nWelfare: {welfare.mean()}({welfare.std()})[{welfare.min()}, {welfare.max()}]')
    # print(p.state)
    assert len(u1) > 0


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_agent_step_limit(init_genius):
    domain = load_genius_domain_from_folder(DOMAIN_FOLDER)
    issues = domain.issues
    p = domain.make_session(
        [
            GeniusNegotiator(
                java_class_name="agents.anac.y2015.Atlas3.Atlas3",
                domain_file_name=DOMAIN_FILE,
                utility_file_name=UTIL1,
            ),
            GeniusNegotiator(
                java_class_name="agents.anac.y2015.AgentX.AgentX",
                domain_file_name=DOMAIN_FILE,
                utility_file_name=UTIL2,
            ),
        ],
        time_limit=None,
        n_steps=20,
    )
    issue_list = [f"{k}:{v}" for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)', 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo', 'Festivals', 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.', 'Monuments', 'Special streets', 'Palace', 'Landscape and nature']",
    ]
    atlas3, agentx = p.negotiators[0], p.negotiators[1]
    p.run()
    # print(f'{len(p.history)} bids exchanged')
    u1 = np.array([float(atlas3.ufun(s.current_offer)) for s in p.history])
    u2 = np.array([float(agentx.ufun(s.current_offer)) for s in p.history])
    u1 + u2
    # print(
    #     f'Negotiator 1: {u1.mean()}({u1.std()})[{u1.min()}, {u1.max()}]\nNegotiator 2: {u2.mean()}({u2.std()})[{u1.min()}, {u1.max()}]'
    #     f'\nWelfare: {welfare.mean()}({welfare.std()})[{welfare.min()}, {welfare.max()}]')
    # print(p.state)


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_agent_step_long_session(init_genius):
    a1 = GeniusNegotiator(
        java_class_name="agents.anac.y2015.Atlas3.Atlas3",
        domain_file_name=DOMAIN_FILE,
        utility_file_name=UTIL1,
    )
    a2 = GeniusNegotiator(
        java_class_name="agents.anac.y2015.Atlas3.Atlas3",
        domain_file_name=DOMAIN_FILE,
        utility_file_name=UTIL2,
    )
    domain = load_genius_domain_from_folder(DOMAIN_FOLDER).normalize()
    issues = domain.issues
    p = domain.make_session([a1, a2], time_limit=None, n_steps=20)
    issue_list = [f"{k}:{v}" for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)', 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo', 'Festivals', 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.', 'Monuments', 'Special streets', 'Palace', 'Landscape and nature']",
    ]
    p.add(a1)
    p.add(a2)
    p.run()
    # print(f'{len(p.history)} bids exchanged')
    u1 = np.array([float(a1._preferences(s.current_offer)) for s in p.history])
    u2 = np.array([float(a2._preferences(s.current_offer)) for s in p.history])
    u1 + u2
    # print(
    #     f'Negotiator 1: {u1.mean()}({u1.std()})[{u1.min()}, {u1.max()}]\nNegotiator 2: {u2.mean()}({u2.std()})[{u1.min()}, {u1.max()}]'
    #     f'\nWelfare: {welfare.mean()}({welfare.std()})[{welfare.min()}, {welfare.max()}]')
    # print(p.state)
    assert len(u1) > 4


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_agent_same_utility_with_normalization():
    from negmas import load_genius_domain_from_folder

    domain = load_genius_domain_from_folder(DOMAIN_FOLDER).normalize()
    issues = domain.issues
    p = domain.make_session(time_limit=300, n_steps=None)
    assert p is not None, "Could not create a mechanism"
    issue_list = [f"{k}:{v}" for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)'"
        ", 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo'"
        ", 'Festivals', 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.'"
        ", 'Monuments', 'Special streets', 'Palace', 'Landscape and nature']",
    ]
    a1 = AgentK(preferences=domain.ufuns[0])
    a2 = Atlas3(preferences=domain.ufuns[0])
    p.add(a1)
    p.add(a2)
    final = p.run()
    u1 = [float(a1.preferences(o)) for o in p.offers]
    u2 = [float(a2.preferences(o)) for o in p.offers]
    assert len(u1) >= 1 or len(u2) >= 1
    u1, u2 = a1.ufun(final.agreement), a2.ufun(final.agreement)
    welfare = u1 + u2
    assert p.state.agreement is not None
    assert p.state.broken is False
    assert welfare > 1


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_agent_same_utility():
    from negmas import load_genius_domain_from_folder

    domain = load_genius_domain_from_folder(DOMAIN_FOLDER)
    issues = domain.issues
    p = domain.make_session(time_limit=300, n_steps=None)
    assert p is not None, "Could not create a mechanism"
    issue_list = [f"{k}:{v}" for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)'"
        ", 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo'"
        ", 'Festivals', 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.'"
        ", 'Monuments', 'Special streets', 'Palace', 'Landscape and nature']",
    ]
    a1 = AgentK(preferences=domain.ufuns[0])
    a2 = Atlas3(preferences=domain.ufuns[0])
    p.add(a1)
    p.add(a2)
    final = p.run()
    u1 = [float(a1.preferences(o)) for o in p.offers]
    u2 = [float(a2.preferences(o)) for o in p.offers]
    assert len(u1) >= 1 or len(u2) >= 1
    u1, u2 = a1.ufun(final.agreement), a2.ufun(final.agreement)
    welfare = u1 + u2
    assert p.state.agreement is not None
    assert p.state.broken is False
    assert welfare > 1


class TestGeniusAgentSessions:
    def prepare(
        self,
        utils=(0, 0),
        single_issue=True,
    ):

        base_folder = pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/Laptop"
        )

        domain = load_genius_domain_from_folder(base_folder).normalize()
        if single_issue:
            domain = domain.to_single_issue(stringify=True)
        atlas = GeniusNegotiator(
            java_class_name="agents.anac.y2015.Atlas3.Atlas3",
            preferences=domain.ufuns[utils[0]],
        )
        # agentx = GeniusNegotiator.random_negotiator(
        agentx = GeniusNegotiator(
            java_class_name="agents.anac.y2015.AgentX.AgentX",
            preferences=domain.ufuns[utils[1]],
        )
        neg = domain.make_session(n_steps=100)
        neg.add(atlas)
        neg.add(agentx)
        return neg

    @pytest.mark.skipif(
        condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
        reason="No Genius Bridge, skipping genius-agent tests",
    )
    def test_genius_agents_can_run_on_converted_single_issue_preferences1(
        self, init_genius
    ):
        neg = self.prepare(utils=(0, 0), single_issue=True)
        best = neg.pareto_frontier(sort_by_welfare=True)[0][0]
        assert isclose(best[0], 1.0, abs_tol=0.001) and isclose(
            best[1], 1.0, abs_tol=0.001
        )
        neg.run()
        assert neg.agreement is not None
        assert len(neg.history) <= 3
        assert neg.agreement == ("v18",)

    @pytest.mark.skipif(
        condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
        reason="No Genius Bridge, skipping genius-agent tests",
    )
    def test_genius_agents_can_run_on_converted_single_issue_preferences2(
        self, init_genius
    ):
        neg = self.prepare(utils=(1, 1), single_issue=True)
        best = neg.pareto_frontier(sort_by_welfare=True)[0][0]
        assert isclose(best[0], 1.0, abs_tol=0.001) and isclose(
            best[1], 1.0, abs_tol=0.001
        )
        neg.run()
        assert neg.agreement is not None
        assert len(neg.history) <= 3
        assert neg.agreement == ("v12",)

    @pytest.mark.skipif(
        condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
        reason="No Genius Bridge, skipping genius-agent tests",
    )
    def test_genius_agents_can_run_on_converted_single_issue(self, init_genius):
        neg = self.prepare(utils=(0, 1), single_issue=True)
        front = neg.pareto_frontier(sort_by_welfare=True)[0]
        assert_almost_equal(
            front,
            [
                (0.7715, 0.845),
                (0.577, 1.0),
                (1.0, 0.5136),
                (0.805, 0.668),
            ],
            decimal=2,
        )

        neg.run()
        assert len(neg.history) >= 1
        assert neg.agreement is not None

    @pytest.mark.skipif(
        condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
        reason="No Genius Bridge, skipping genius-agent tests",
    )
    def test_genius_agents_can_run_on_converted_multiple_issues(self, init_genius):
        neg = self.prepare(utils=(0, 0), single_issue=False)
        frontier = neg.pareto_frontier(sort_by_welfare=True)[0]
        true_frontier = [(1.0, 1.0)]
        assert len(frontier) == len(true_frontier)
        for a, b in zip(frontier, true_frontier):
            assert abs(a[0] - b[0]) < 1 and abs(a[1] - b[1]) < 1
        # neg.set_sync_call(True)
        neg.run()
        assert len(neg.history) < 3
        assert neg.agreement is not None
        assert neg.agreement == ("HP", "60 Gb", "19'' LCD")

    # @pytest.mark.skipif(
    #     condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    #     reason="No Genius Bridge, skipping genius-agent tests",
    # )
    # def test_genius_agents_can_run_on_converted_multiple_issues_no_names(
    #     self, init_genius
    # ):
    #     neg = self.prepare(
    #         utils=(0, 0),
    #         single_issue=False,
    #         keep_issue_names=True,
    #         keep_value_names=True,
    #     )
    #     frontier = neg.pareto_frontier(sort_by_welfare=True)[0]
    #     true_frontier = [(1.0, 1.0)]
    #     assert len(frontier) == len(true_frontier)
    #     for a, b in zip(frontier, true_frontier):
    #         assert abs(a[0] - b[0]) < 1 and abs(a[1] - b[1]) < 1
    #
    #     state = neg.run()
    #     assert len(neg.history) < 4, len(neg.history)
    #     assert neg.agreement is not None
    #     # assert neg.agreement == {
    #     #     "External Monitor": "19'' LCD",
    #     #     "Harddisk": "60 Gb",
    #     #     "Laptop": "HP",
    #     # }
    #     assert (neg.agreement.values()) == ("HP", "60 Gb", "19'' LCD"), neg.agreement

    @pytest.mark.skipif(
        condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
        reason="No Genius Bridge, skipping genius-agent tests",
    )
    def test_genius_agent_example(self, init_genius):
        agent_name1 = "agents.anac.y2015.Atlas3.Atlas3"
        agent_name2 = "agents.anac.y2015.Atlas3.Atlas3"
        single_issue = False

        base_folder = pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/Laptop"
        )
        domain = load_genius_domain_from_folder(base_folder)
        if single_issue:
            domain = domain.to_single_issue()
        # atlas = GeniusNegotiator.random_negotiator(
        atlas = GeniusNegotiator(
            java_class_name=agent_name1,
            domain_file_name=domain.agenda.name,
            utility_file_name=domain.ufuns[0].name,
        )
        agentx = GeniusNegotiator(
            java_class_name=agent_name2,
            domain_file_name=domain.agenda.name,
            utility_file_name=domain.ufuns[1].name,
        )
        neg = domain.make_session(n_steps=100)
        neg.add(atlas)
        neg.add(agentx)


if __name__ == "__main__":
    pytest.main(args=[__file__])
